/* -*-c++-*- source - Open Producer - Copyright (C) 2002 Don Burns
 * Distributed under the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE (LGPL)
 * as published by the Free Software Foundation.
 */

#include <Producer/RenderSurface>

#ifdef _X11_IMPLEMENTATION
#include <stdio.h>
#include <unistd.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xmd.h>


using namespace Producer; 

// HP, SGI and SUN don't implement glXGetProcAddressARB...
#if defined (__sgi) || defined (sun) || defined( __hpux ) 
#include <dlfcn.h>
void (*glXGetProcAddressARB(const GLubyte *procName))(void)
{
    static void *handle = dlopen( (const char *)0L, RTLD_LAZY );
    return (void (*)(void))dlsym(handle,(const char *)procName);
}
#endif

#if (defined(__APPLE__) || defined(macintosh)) & defined (GLX_VERSION_1_4)
#define glXGetProcAddressARB glXGetProcAddress
#endif


bool RenderSurface::realize( VisualChooser *vc, GLContext shared_context )
{
    if( _realized ) return _realized;

    if( vc != NULL )
        _visualChooser = vc;
    _sharedGlcontext = shared_context;
    _init();

    if( _useConfigEventThread )
    {
        _threadReady = new RefBarrier(2);
        startThread();
        _threadReady->block();
    }

    XMapWindow( _dpy, _win );
    Window temp = _win;
    XSetWMColormapWindows( _dpy,  _win, &temp, 1);
    makeCurrent();
#if 0
    // NVidia 1.0-4496 driver breaks (once again), glXGetVideoSync.  This problem was seen
    // earlier on laptops (Geforce4Go), but seems to be present on desktops as
    // wells now.  A bug will be filed with NVidia again.
    //
    // glXWaitVideoSyncSGI is used for synchronizing the CPU bound process to
    // video retrace signal and is ultimately not necessary, but would provide
    // consistent frame rates in the CPU if it was present.
    //
    // Note that testVSync hangs if the window has not been mapped.  If the window
    // has been mapped, glXGetVideoSync() fails
    testVSync();
#endif

    _realized = true;
 
    std::vector <Producer::ref_ptr<Callback> >::iterator p;
    for( p = _realizeCallbacks.begin(); p != _realizeCallbacks.end(); p++ )
    {
        if( (*p).valid() )
            (*(*p).get())( *this );
    }
    _realizeCallbacks.clear();

    _realizeBlock->release();

    return _realized;
}
 
bool RenderSurface::_init()
{
    char dpyname[128];
    bool _dpy_created_locally = false;

    if( _hostname.empty() )
        sprintf( dpyname, ":%d.%d",  _displayNum, _screen );
    else
        sprintf( dpyname, "%s:%d.%d", _hostname.c_str(), _displayNum, _screen );

    // Set Display
    if( _dpy == NULL )
    {
        if( (_dpy = XOpenDisplay( dpyname )) == (Display *)0L )
        {
            std::cerr << "RenderSurface::realize() - unable to open display \""
                <<XDisplayName( dpyname) << "\"" << std::endl;
            return false;
        }
        _dpy_created_locally = true;
    }

        // Query for GLX extension
    int erb, evb;
    if( glXQueryExtension( _dpy, &erb, &evb)  == False )
    {
        std::cerr << "RenderSurface::realize() " << 
                    XDisplayName( dpyname) <<
                    " has no GLX extension." << std::endl;
         if( _dpy_created_locally )
         {
             XCloseDisplay( _dpy );
             _dpy = 0;
         }
         return false;
    }
    
    if( _visualInfo == NULL )
    {
    // If Window was already defined, lets grab that visual
        if( _win != 0 )
        {
            XWindowAttributes watt;
            XGetWindowAttributes( _dpy, _win, &watt );
            XVisualInfo temp;
            temp.visualid = XVisualIDFromVisual(watt.visual);
            int n;
            _visualInfo = XGetVisualInfo( _dpy, VisualIDMask, &temp, &n );
        }
        else
        {
            if( _visualChooser == NULL )
            {
                ref_ptr<VisualChooser> lvc = new VisualChooser;
                lvc->setSimpleConfiguration();
                _visualInfo = lvc->choose(_dpy, _screen);
            }
            else
            {
                _visualInfo = _visualChooser->choose(_dpy,_screen);
            }
        }
    }
    
    if( _visualInfo == NULL )
    {
        if( _dpy_created_locally )
        {
            XCloseDisplay( _dpy );
            _dpy = 0;
        }
        return false;
    }
        
    if((_glcontext = glXCreateContext( _dpy, _visualInfo, _sharedGlcontext, GL_TRUE ))==(GLXContext) 0L)
    {
        std::cerr << "RenderSurface:realize() - Can't create GLX Context." << std::endl;
        if( _dpy_created_locally )
        {
            XCloseDisplay( _dpy );
            _dpy = 0L;
        }
        return false;
    }


    if( _win == 0 )
    {
        if( _parent == (Window)0L )
        {
            _parent = RootWindow( _dpy, _screen );
            _mayFullScreen = true;
        }
        else
        {
            _mayFullScreen = false;
        }

        XWindowAttributes watt;
        XGetWindowAttributes( _dpy, _parent, &watt );
        _parentWindowHeight = watt.height;

        if( _windowWidth == 0xFFFFFFFF || (_mayFullScreen && _isFullScreen))
        {
            _windowWidth = watt.width;
            _decorations = false;
        }

        if( _windowHeight == 0xFFFFFFFF || (_mayFullScreen && _isFullScreen))
        {
            _windowHeight = watt.height;
            _decorations = false;
        }

        XSetWindowAttributes swatt;
        swatt.colormap = XCreateColormap( _dpy, RootWindow( _dpy, _screen ),
                                        _visualInfo->visual, AllocNone);
    //swatt.colormap = DefaultColormap( _dpy, 0 );
        swatt.background_pixel = 0;
        swatt.border_pixel = 0;
        swatt.event_mask =  0;
        unsigned long mask = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap;
        //swatt.override_redirect = true;
        //unsigned long mask = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap | CWOverrideRedirect;
        _win = XCreateWindow( _dpy, _parent,
                              _windowX,
                              _parentWindowHeight - (_windowHeight + _windowY),
                              _windowWidth, _windowHeight, 0,
                              _visualInfo->depth, InputOutput,
                              _visualInfo->visual, mask, &swatt );

        // This positions the window at _windowX, _windowY
        XSizeHints sh;
        sh.flags = 0;
        sh.flags |= USSize;
        sh.flags &= 0x7;
        sh.flags |= USPosition;
        sh.flags &= 0xB;
        sh.x = _windowX;
        sh.y = _parentWindowHeight - (_windowHeight + _windowY);
        sh.width  = _windowWidth;
        sh.height = _windowHeight;
        XSetStandardProperties( _dpy, _win, _windowName.c_str(), _windowName.c_str(), None, 0, 0, &sh);


        // This turns off border decorations
        if( _decorations == false )
            _setBorder(_decorations);

        // Create Null Cursor
        {
            Pixmap pixmap;
            static char buff[2] = {0,0};
            static XColor ncol = {0,0,0,0,DoRed|DoGreen|DoBlue,0};

            pixmap = XCreateBitmapFromData( _dpy, DefaultRootWindow(_dpy), buff, 1, 1);
           _nullCursor = XCreatePixmapCursor( _dpy, pixmap, pixmap, &ncol, &ncol, 0, 0 );
        }
        _setCursor();
    }
    else
    {
        XWindowAttributes watt;
        XGetWindowAttributes( _dpy, _win, &watt );
        _windowX = watt.x;
        _windowY = watt.y;
        _windowWidth = watt.width;
        _windowHeight = watt.height;

        // We can't know if the imposed window may resize or not
        _mayFullScreen = false;
        
        // We must not check for events.  Leave that up to the externally created window
        _checkOwnEvents = false;

        // Set Screen Number
        _screen = -1;
        for( int i = 0; i < ScreenCount(_dpy); i++ )
        {
            if( watt.screen == ScreenOfDisplay(_dpy,i) )
            {
                _screen = i;
                break;
            }
        }
        if( _screen == -1 )
        {
            // A warning might be appropriate here
            _screen = 0;
        }
    }

    _screenWidth = DisplayWidth(_dpy, _screen );
    _screenHeight = DisplayHeight(_dpy, _screen );

    if( _windowHeight == _screenHeight && _windowWidth == _screenWidth )
        _isFullScreen = true;
    else
        _isFullScreen = false;

    if( !_useConfigEventThread )
        XSelectInput( _dpy, _win, ExposureMask | StructureNotifyMask );

    XFlush( _dpy );
    XSync( _dpy, 0 );
    return true;
}

bool RenderSurface::makeCurrent( void ) 
{
    glXMakeCurrent( _dpy, _win, _glcontext );
    return true;
}

bool RenderSurface::makeCurrent( void ) const
{
    glXMakeCurrent( _dpy, _win, _glcontext );
    return true;
}

void RenderSurface::swapBuffers( void )
{
    glXMakeCurrent( _dpy, _win, _glcontext );
    glXSwapBuffers( _dpy, _win );
    if( _checkOwnEvents && _useConfigEventThread == false )
    {
        while( XPending(_dpy) )
           _checkEvents(_dpy);
    }
}


void RenderSurface::sync( int divisor )
{
    // In spite of repeated efforts to get Nvidia and the ARB to 
    // Communicate on this issue, the ARB definition of glXWaitVideoSyncSGI()
    // is ambiguous and NVidia's implementation of it favors the least
    // likely interpretation.
    // So, it is necessary to call GetVideoSync before EVERY WaitVideoSync
    // if you want WaitVideoSync to work correctly when divisor = 1, on
    // NVidia.  On SGI, the penalty is an extra call to GetVideoSync 
    if( __glxGetVideoSyncSGI != NULL && __glxWaitVideoSyncSGI != NULL )
    {
        // This may be fixed now...
        // 5/16/03  Keep this in.  If divisor is anything other than
        // 0 this still does not work
        __glxGetVideoSyncSGI( &_frameCount );
        __glxWaitVideoSyncSGI( divisor, 0, &_frameCount );
    }
    // else do nothing.  Warning already printed in testVSync.
}


void RenderSurface::testVSync( void )
{
    unsigned int frame = 0;
    __glxGetRefreshRateSGI = reinterpret_cast<int (*)(unsigned int *)>
        (glXGetProcAddressARB((GLubyte *)"glXGetRefreshRateSGI" ));
    __glxWaitVideoSyncSGI = reinterpret_cast<int(*)(int,int,unsigned int *)>
        (glXGetProcAddressARB((GLubyte *)"glXWaitVideoSyncSGI" ));
    __glxGetVideoSyncSGI = reinterpret_cast<int(*)(unsigned int *)>
        (glXGetProcAddressARB((GLubyte * )"glXGetVideoSyncSGI"));

    if( __glxGetRefreshRateSGI != NULL )
    __glxGetRefreshRateSGI(&_refreshRate);

    // Prior reference to a bug here in the nVidia drivers is moot.
    // Bug now seems to be fixed.
    if( __glxGetVideoSyncSGI != NULL && __glxWaitVideoSyncSGI != NULL )
    {
        __glxGetVideoSyncSGI(&frame);
        if( frame != 0 )
        {
            __glxWaitVideoSyncSGI( 1, 0, &frame );
            _frameCount = frame;
            __glxWaitVideoSyncSGI( 1, 0, &frame );
            __glxWaitVideoSyncSGI( 1, 0, &frame );

            if( frame != (_frameCount + 2) )
                std::cerr << "Warning... Video Sync not working correctly\n";
        }
        else
            std::cerr << "Warning... Video Sync not working correctly (glXGetVideoSyncSGI() failed)\n";
    }
    else
        std::cerr << "Warning... Video Sync capability not available\n";

    _frameCount = frame;
}

bool RenderSurface::_checkEvents( Display *dpy )
{
    bool ret = true;
    XEvent ev;
    XNextEvent( dpy, &ev );
// Window manager will reparent and reset the x,y location relative
// to the window manager's border window.  On a reparent, we set 
// this flag.  Seems to work but .....
static bool _ignore_resize_flag_hack = false;
    switch( ev.type )
    {
        case Expose :
          break;

        case ConfigureNotify :
           if( !_isFullScreen && !_ignore_resize_flag_hack )
           {
                _windowWidth = ev.xconfigure.width;
                _windowHeight = ev.xconfigure.height;
                _windowX = ev.xconfigure.x;
                _windowY = DisplayHeight(dpy,_screen) - (ev.xconfigure.y+_windowHeight);

                if( _bindInputRectangleToWindowSize == true )
                    _inputRectangle.set( 0.0, float(_windowWidth), 0.0, float (_windowHeight));

           }
           _ignore_resize_flag_hack = false;
        break;

        case GravityNotify :
        break;

        case MapNotify :
          {
             XWindowAttributes watt;
            do
                XGetWindowAttributes(dpy, _win, &watt );
            while( watt.map_state != IsViewable );

            XSetInputFocus( dpy, _win, RevertToNone, CurrentTime );
            XFlush(dpy); XSync(dpy,0);
          }
            break;

        case UnmapNotify :
            break;

        case ReparentNotify:
_ignore_resize_flag_hack = true;    
            break;

        case DestroyNotify :
            _realized =  false;
            ret = false;
            break;
    }
    return ret;
}

void RenderSurface::run( void )
{
    Display *ldpy;
    char dpyname[128];
    if( _hostname.empty() )
        sprintf( dpyname, ":%d.%d",  _displayNum, _screen );
    else
        sprintf( dpyname, "%s:%d.%d", _hostname.c_str(), _displayNum, _screen );
    if( (ldpy = XOpenDisplay( dpyname )) == (Display *)0L )
    {
        std::cerr << "RenderSurface() : Reconfigure event thread - "
                     "Unable to open display \"" 
                  << XDisplayName( dpyname ) << "\"" << std::endl;
        //return (void *)1;
        return;
    }

    const unsigned long mask = ExposureMask | StructureNotifyMask;
    XSelectInput( ldpy, _win,  mask );

    if( _threadReady.valid() )
        _threadReady->block();

    for( ;; )
    {
#if 0
        fd_set rfds;
        int fd = ConnectionNumber(ldpy);
        FD_ZERO(&rfds);
        FD_SET( fd, &rfds );
        if( select(fd+1,&rfds,0,0,0) < 0 )
            testCancel();
#endif
        do
        {
            _checkEvents(ldpy);
            testCancel();
        } while(XPending(ldpy));
    }

    XCloseDisplay( ldpy );
}



void RenderSurface::fullScreen( bool flag )
{
    if( flag == _isFullScreen ) return;

    if( flag )
    {
        if( _mayFullScreen == false || !_realized ) 
        {
            if( !_realized )
                std::cerr << "RenderSurface::fullScreen() - Cannot set full screen before realization\n";
            return;
        }

        _isFullScreen = true;
        _setBorder(false);
        XFlush(_dpy); XSync(_dpy,0);
        XUnmapWindow(_dpy, _win );
        XFlush(_dpy); XSync(_dpy,0);
        usleep(100000); // sleep to try and give the window manager a chance to resize properly - accounting for bugggy window manager..
        XMoveResizeWindow( _dpy, _win, 0, 0, DisplayWidth(_dpy,_screen), DisplayHeight(_dpy,_screen) );
        XFlush(_dpy); XSync(_dpy,0);
        XMapWindow(_dpy, _win );
        XFlush(_dpy); XSync(_dpy,0);
        XSetInputFocus(_dpy, _win, RevertToNone, CurrentTime );
//    XFlush(_dpy); XSync(_dpy,0);
    }
    else
    {
        _isFullScreen = false;
        _resizeWindow();
        _setBorder(_decorations);
        XFlush(_dpy); XSync(_dpy,0);
    }
}

void RenderSurface::_resizeWindow()
{
    XFlush(_dpy); XSync(_dpy,0);
    XUnmapWindow(_dpy,_win);
    XFlush(_dpy); XSync(_dpy,0);
    usleep(100000); // sleep to try and give the window manager a chance to resize properly - accounting for bugggy window manager..
    XMoveResizeWindow( _dpy, _win, 
                    _windowX, 
                    DisplayHeight(_dpy,_screen) - (_windowY + _windowHeight), 
                    _windowWidth, _windowHeight );
    XFlush(_dpy); XSync(_dpy,0);
    XMapWindow(_dpy, _win);
    XFlush(_dpy); XSync(_dpy,0);
}

void RenderSurface::_setBorder( bool flag )
{
    Atom atom;
    if( (atom = XInternAtom( _dpy, "_MOTIF_WM_HINTS", 0 )) != None )
    {
// Hack for sending 64 bit atom to Xserver
#if defined( _MIPS_SIM) && (_MIPS_SIM == _MIPS_SIM_ABI64)
        struct {
            CARD32 flags0;
            CARD32 flags1;
            CARD32 functions0;
            CARD32 functions1;
            CARD32 decorations0;
            CARD32 decorations1;
            INT32 input_mode0;
            INT32 input_mode1;
            CARD32 status0;
            CARD32 status1;
        } wmHints;

        wmHints.flags1       = (1L << 1);
        wmHints.functions1   = 0;
        wmHints.decorations1 = flag;
        wmHints.input_mode1  = 0;
#else
        struct {
            CARD32 flags;
            CARD32 functions;
            CARD32 decorations;
            INT32 input_mode;
            CARD32 status;
        } wmHints;

        wmHints.flags       = (1L << 1);
        wmHints.functions   = 0;
        wmHints.decorations = flag;
        wmHints.input_mode  = 0;
#endif
        XChangeProperty( _dpy, _win, atom, atom, 32, PropModeReplace,
                (unsigned char *)&wmHints,  5 );

        XFlush(_dpy);
        XSync(_dpy,0);
    }
    else
        std::cerr << 
         "RenderSurface::setBorder(" << flag << ") - Warning : Couldn't change decorations." << std::endl;
}

void RenderSurface::_setCursor()
{
    if( _useCursor == false )
        XDefineCursor( _dpy, _win, _nullCursor );
    else
        XUndefineCursor( _dpy, _win );
    XFlush(_dpy);
}

void RenderSurface::_positionPointer( int x, int y )
{
    if( !_realized )
        return;

    if( _isFullScreen )
        y = _screenHeight - y;
    else
        y = _windowHeight - y;
    XWarpPointer( _dpy, 
                    None,
                    _win, 
                    0, 0, 0, 0,
                    x, y );
    XFlush(_dpy);
    XSync(_dpy,0);
}

void RenderSurface::getScreenSize( unsigned int &width, unsigned int &height )  const
{
    if( _realized )
    {
        width = _screenWidth;
        height = _screenHeight;
    }
    else
    {
        Display *dpy = XOpenDisplay("");
        width = DisplayWidth(dpy, _screen);
        height = DisplayHeight(dpy,_screen);
        XCloseDisplay(dpy);
    }
}

void RenderSurface::getScreenSize( unsigned int &width, unsigned int &height )  
{
    if( _realized )
    {
        width = _screenWidth;
        height = _screenHeight;
    }
    else
    {
        Display *dpy = XOpenDisplay("");
        width = DisplayWidth(dpy, _screen);
        height = DisplayHeight(dpy,_screen);
        XCloseDisplay(dpy);
    }
}

void RenderSurface::mapWindow()
{
    if( _realized && _parent == RootWindow( _dpy, _screen ) )
    {
        XMapWindow( _dpy, _win );
        XFlush(_dpy);
        XSync(_dpy,0);
    }
}

void RenderSurface::unmapWindow()
{
    if( _parent == RootWindow( _dpy, _screen ) )
    {
        XUnmapWindow( _dpy, _win );
        XFlush(_dpy);
        XSync(_dpy,0);
    }
}


void RenderSurface::_fini()
{
    if( _realized )
    {
        //glXDestroyContext( _dpy, _glcontext );
        XDestroyWindow(_dpy, _win );
        XFlush( _dpy );
        XSync(_dpy,0);
    }
    if( _visualInfo != NULL )
        XFree(_visualInfo);
}
#endif

